use std::collections::HashMap;

use crate::{
    parser::{Instruction, Value},
    Result,
};

type Function = fn(&mut Assembler, usize);

pub struct Assembler {
    functions: HashMap<String, Function>,
    stack: Vec<String>,
    assembled: Vec<u8>,
}

trait Assemble {
    fn assemble(self, assembler: &mut Assembler) -> Result<()>;
}

impl Assemble for Value {
    fn assemble(self, assembler: &mut Assembler) -> Result<()> {
        match self {
            Self::Instruction(instruction) => instruction.assemble(assembler),
            Value::IntegerLiteral(v) => {
                let bytes = v.to_le_bytes();

                assembler.push(v.to_string());

                assembler.add_instruction(&[0x68]);
                assembler.add_instruction(&bytes);

                Ok(())
            }
            _ => todo!(),
        }
    }
}

impl Assemble for Instruction {
    fn assemble(self, assembler: &mut Assembler) -> Result<()> {
        let n_arguments = self.arguments.len();

        for argument in self.arguments {
            argument.assemble(assembler)?;
        }

        let function = match assembler.functions.get(&self.name) {
            Some(v) => v,
            None => return Err(crate::Error::UndefinedIdentifier(self.name)),
        };

        function(assembler, n_arguments);

        Ok(())
    }
}

impl Assembler {
    pub fn assemble(instruction: Instruction) -> Result<Vec<u8>> {
        let mut assembler = Self {
            functions: Self::default_functions(),
            stack: Vec::new(),
            assembled: Vec::new(),
        };
        instruction.assemble(&mut assembler)?;
        Ok(assembler.assembled)
    }

    fn add_instruction(&mut self, instruction: &[u8]) {
        self.assembled.extend_from_slice(instruction);
    }

    fn push(&mut self, identifier: String) {
        self.stack.push(identifier);
    }

    fn pop(&mut self) {
        self.stack.pop();
    }

    fn get_stack_location(&self, identifier: &str) -> Option<u32> {
        self.stack
            .iter()
            .position(|x| x == identifier)
            .map(|x| x as u32)
    }

    fn default_functions() -> HashMap<String, Function> {
        let mut map: HashMap<String, Function> = HashMap::new();

        map.insert("+".to_string(), |assembler, n_arguments| {
            if n_arguments != 2 {
                todo!()
            }

            assembler.add_instruction(&[0x58]); // Pop EAX
            assembler.pop();

            assembler.add_instruction(&[0x03, 0x04, 0x24]); // add (%esp), %eax

            assembler.add_instruction(&[0x5B]); // Pop EBX
            assembler.pop();

            assembler.add_instruction(&[0x50]); // Push EAX
            assembler.push("".to_string());
        });

        map
    }
}
