use std::str::FromStr;

use thiserror::Error;

#[derive(Debug)]
pub enum TokenType {
    OpenParantheses,
    ClosedParantheses,
    Integer(usize),
    String(String),
    Identifier(String),
}

#[derive(Debug)]
pub struct Token {
    row: usize,
    col: usize,

    token_type: TokenType,
}

#[derive(Debug)]
pub struct TokenString {
    tokens: Vec<Token>,
}

#[derive(Debug)]
struct UnfinishedTokenString {
    position: usize,
    input: Vec<u8>,
    tokens: Vec<Token>,
    eof: bool,

    row: usize,
    col: usize,
}

#[derive(Error, Debug)]
#[error("Syntax error at: {row}:{col}. {message}")]
pub struct LexerError {
    row: usize,
    col: usize,
    message: String,
}

impl FromStr for TokenString {
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let uts = UnfinishedTokenString {
            position: 0,
            input: s.to_string().as_bytes().to_vec(),
            tokens: Vec::new(),
            eof: false,

            row: 0,
            col: 1,
        };

        uts.lex_file()
    }

    type Err = LexerError;
}

impl From<UnfinishedTokenString> for TokenString {
    fn from(value: UnfinishedTokenString) -> Self {
        Self {
            tokens: value.tokens,
        }
    }
}

trait Lexer {
    fn lex_file(self) -> Result<Self::TokenString, Self::Err>;

    type TokenString;
    type Err;
}

impl Lexer for UnfinishedTokenString {
    fn lex_file(mut self) -> Result<Self::TokenString, Self::Err> {
        self.skip_whitespace();

        while !self.eof {
            use TokenType as TT;

            let token_type = match self.input[self.position] as char {
                '(' => Ok(TT::OpenParantheses),
                ')' => Ok(TT::ClosedParantheses),
                'a'..='z' | 'A'..='Z' => Ok(TT::Identifier(self.lex_identifier())),
                '0'..='9' | '-' => Ok(TT::Integer(self.lex_integer())),
                v => Err(format!("Expected new token, found: {v}")),
            };

            let token = match token_type {
                Ok(v) => Token {
                    col: self.col,
                    row: self.row,
                    token_type: v,
                },
                Err(e) => {
                    return Err(LexerError {
                        col: self.col,
                        row: self.row,
                        message: e,
                    })
                }
            };

            self.tokens.push(token);

            self.next_char();
            self.skip_whitespace();
        }

        return Ok(self.into());
    }

    type TokenString = TokenString;
    type Err = LexerError;
}

impl UnfinishedTokenString {
    fn skip_whitespace(&mut self) {
        while !self.eof && self.input[self.position].is_ascii_whitespace() {
            self.next_char();
        }
    }

    fn next_char(&mut self) {
        self.position += 1;
        if self.position == self.input.len() {
            self.eof = true;
            return;
        }
        if self.input[self.position] == b'\n' {
            self.row += 1;
            self.col = 0;
        } else {
            self.col += 1;
        }
    }

    fn lex_identifier(&mut self) -> String {
        todo!()
    }

    fn lex_integer(&mut self) -> usize {
        todo!()
    }
}
